c *****************************************************************************
c ***   Dec mpp12000 Fortran Random Numbers Generator                       ***
c ***   =============================================                       ***
c ***    Provides uniformly distributed random numbers using a combination  ***
c ***    of lagged Fibonaci and arithmetic sequence generators. It assumes  ***
c ***    that the desired geometry of target arrays is 2 dimensional and    ***
c ***    of sizes 128x64, but can be changed to any target dimension/size.  ***
c ***                                                                       ***
c ***   By J. Apostolakis, M. Guagnelli, E. Marinari                        ***
c ***   --------------------------------------------                        ***
c ***   Syracuse University -- v1.0 -- March, 21, 1992                      ***
c *****************************************************************************
c ***   Based on universal random number generator for sequential computers ***
c ***   of Marsaglia, Zaman and Tsang, Stat. Prob. Lett. 9 (1990) 35.       ***
c ***   as presented in F. James, Comp. Phys. Comm. 60 (1990) 329, it is    ***
c ***   the generator "RANMAR" p.339                                        ***
c ***   With comments, or for corrections or updates contact                ***
c ***                                   japost@nova.npac.syr.edu            ***
c *****************************************************************************
	program mp_random
 	implicit none

 	real, array(0:127,0:63) :: arandno 
        integer istate, ti, tj
        real    sum

        interface
            function ranmar( jstate )
            real, array(0:127,0:63) :: ranmar
            integer jstate
            end function 
        end interface

	integer      seed, seed1, seed2, seed_max 
	parameter (seed_max = 30082)

 	print *, 'Enter Random Seed '
	read  *, seed 

	seed1 = seed / seed_max
	seed2 = seed - (seed_max * seed1)

	call init_random ( seed1, seed2 )
	arandno = ranmar( istate )

        sum = 0.0
	do ti = 0, 127
	    do tj = 0, 63
                sum = sum + arandno( ti, tj )
            enddo
        enddo
        print *, ' The sum of all the 1st random nos is ', sum

        stop
	end

c ------------------------------------------------------------------

c **************************************************************************
c ***  SUBROUTINE init_random ( seed1, seed2 )
c ***   Accepts two seeds and initialises all working parts of the 
c ***   random number generator(RNG). Uses a simple sequential RNG (#1) to
c ***   seed a parallel version of MZT's initialising RNG (#2), which in
c ***   turn is used to seed the lagged Fibonaci part of the overall RNG.
c ***   The seeds must be in the range [0,31328] and [0,30081] respectively 
c **************************************************************************
	subroutine init_random ( seed1, seed2 )
	implicit none

c *** Parameters of RNG No 1 , of the multiplicative linear congruent    ***
c *** type which is used to provide indepedent starting points for       ***
c *** the seeding of the initialising parallel RNG.                      ***
c **************************************************************************
	integer    ia, ic, im, x1_max, x2_max
	parameter (ia = 967)
	parameter (ic = 3041)
	parameter (im = 14406)
c *** The upper limits on the seeds for the parallel initialising        ***
c ***  generator (Rng no 2).      					 ***
c **************************************************************************
	parameter (x1_max = 31328)
	parameter (x2_max = 30081)

	integer    i97, j97, ii, jj, Seed1, Seed2
	integer,   array(0:127,0:63)    :: se1, se2, i, j, k, l, m
	real       c, cd, cm
	real,      array(0:127,0:63,97) :: u
	real,      array(0:127,0:63)    :: s, t

c **************************************************************************
c **  The variables in these common statements specify the state of the
c **  parallel random number generator and must all be saved and restored
c **  in order to continue in a later run from a certain point in a sequence.
c **************************************************************************
	common /mp_urandom/   u
  	common /mp_numbers/   c, cd, cm, i97, j97
	
CMPF ONDPU mp_urandom
CMPF ONDPU se1, se2, i, j, k, l, m
CMPF MAP   u   ( XBITS, YBITS, MEMORY )
CMPF MAP   se1 ( XBITS, YBITS )
CMPF MAP   se2 ( XBITS, YBITS )
CMPF MAP   i   ( XBITS, YBITS )
CMPF MAP   j   ( XBITS, YBITS )
CMPF MAP   k   ( XBITS, YBITS )
CMPF MAP   l   ( XBITS, YBITS )
CMPF MAP   m   ( XBITS, YBITS )
CMPF MAP   s   ( XBITS, YBITS )
CMPF MAP   t   ( XBITS, YBITS )

c **************************************************************************
c ** Loop over processors, getting new seeds for each processor.
c **************************************************************************

	do ii = 0, 127
	  do jj = 0, 63
	    seed1 = mod((seed1 * ia + ic), im)
	    seed2 = mod((seed2 * ia + ic), im)
	    se1(ii,jj) = ((x1_max + 1) * seed1) / im
	    se2(ii,jj) = ((x2_max + 1) * seed2) / im
	  enddo
	enddo

c **************************************************************************
c *** Having initialized the seeds arrays se1 and se2, we use them      ***
c *** to initialize the working array u of the Fibonaci part of our     ***
c *** random number generator, in parallel.                  		***
c *************************************************************************

	i = mod(se1/177, 177) + 2
	j = mod(se1    , 177) + 2
        k = mod(se2/169, 178) + 1
	l = mod(se2,     169)

	do ii = 1, 97
	  s = 0.0
	  t = 0.5
	  do jj = 1, 24
	    m = mod(mod(i*j, 179)*k, 179)
	    i = j
	    j = k
 	    k = m
	    l = mod(53*l+1, 169)
	    where ( mod(l*m, 64) .ge. 32 )
	      s = s + t
	    endwhere
	    t = 0.5 * t
	  enddo
	  u(:,:,ii) = s
	enddo
	c   = 362436.0   / 16777216.0
	cd  = 7654321.0  / 16777216.0
	cm  = 16777213.0 / 16777216.0
	i97 = 97
	j97 = 33

	return
	end

c --------------------------------------------------------------------------

c **************************************************************************
c ***  FUNCTION pranmar( seed )                                          ***
c ***   Returns one random number in [0,1) on every processor of an 8k   ***
c ***   machine configured as a 128x64 array                             ***
c **************************************************************************
        function ranmar( state )
        implicit none
 	real, array(0:127,0:63) :: ranmar
        integer state
c       this integer is unused, but required to make function compile.

	integer  i97, j97, ii, jj
	real     c, cd, cm
	real,    array(0:127,0:63,97) :: u
	real,    array(0:127,0:63)    :: r

	common /mp_urandom/   u
	common /mp_rrandom/   r
  	common /mp_numbers/   c, cd, cm, i97, j97
	
CMPF ONDPU mp_urandom
CMPF ONDPU mp_rrandom
CMPF MAP   u   ( XBITS, YBITS, MEMORY )
CMPF MAP   r   ( XBITS, YBITS )

	r = u(:,:,i97) - u(:,:,j97)
	where ( r .lt. 0.0 ) r = r + 1.0
	u(:,:,i97) = r
	i97 = i97 - 1
	if ( i97 .eq. 0 ) i97 = 97
	j97 = j97 - 1
	if ( j97 .eq. 0 ) j97 = 97
	c = c - cd
	if ( c .lt. 0.0 ) c = c + cm
	r = r - c
	where ( r .lt. 0.0 ) r = r + 1.0

        ranmar = r
	return
	end

c -----------------------------------------------------------------------

